from .logger import Logger

from atelier_fabrique.message import Message, MessageType, Header
from atelier_fabrique.constants import API_IN_TOPIC, MONITOR_IN_TOPIC, outgoing_fields
from atelier_fabrique.constants import MessageFieldsConstants as fld

from time import time

import sys
import traceback

from collections import deque


METRIC_FIELDS = [fld.error, fld.alerts, fld.metrics, fld.to_detector, fld.named_features] #hardcoded body fields for metrics and alerts

   
class Producer(object):
    def __init__(self, store):
        self.store = store
        #print(self.store)
        
    def send(self, topic, body, kafka_key=None):
        #print(f'!send to topic {topic}')
        self.store.produce_kaf_format(topic, body, kafka_key)

class MessageBus(object):
    def __init__(self, config_dict):
        self.metric_fields = METRIC_FIELDS
        self.config = config_dict
        
        self.store = self.config['message_bus']['servers']
        
        self.client_id = self.config["message_bus"]["client_id"] 
        self.pipeline_ver = self.config["message_bus"].get("pipeline_ver") 
        self.pipeline_name = self.config["message_bus"].get("pipeline_name")

        #logger
        self.logger = Logger.make_logger(self.client_id)
        
        #producer+consumer                 
        to_list = lambda x: x if type(x)==list else ([x,] if x else []) #converts all types to list type        
        #1. лист входных топиков или один топик - неважно
        self.in_topics = to_list(self.config['message_bus'].get('in_topic')) 
        #2. вЫходной топик для данных, результатов, скоров и т.д.
        self.out_topic = self.config['message_bus'].get('out_topic') 
        #3. топик для ошибок
        self.error_topic = self.config['message_bus'].get('error_topic') 
        #4. топик для мониторинга
        self.monitor_topics = self.config['message_bus'].get('monitor_topics')
        #5. выходные топики для диспетчера
        self.out_topic_dict = self.config['message_bus'].get('out_topic_dict')
        
        self.producer = Producer(self.store)    
    
    def send_message(self, header, body, topic=None, message_type=None, kafka_key=None):        
        self.logger.debug('! send msg with session_id = {} to topic = {}'.format(header.session_id, topic))
        #print('! send msg with session_id = {} to topic = {}'.format(header.session_id, topic))
        if not topic: 
            return 
                
        mes = Message.gen_message_by_header(header, body, message_type, self.client_id)
        try:
            if kafka_key:            
                self.producer.send(topic, mes.serialize(), kafka_key)
            else:
                self.producer.send(topic, mes.serialize())
            
        except TypeError:
            print(f'!Cannot send to {topic}, {type(self).__name__}, {mes.body}')
            raise
        
    def send_data(self, header, body, topic=None):
        '''Отправлялка данных'''
        #print(f"topic = {topic}")
        if not topic:
            topic=self.out_topic
        
        if fld.data in body:
            mes = {}
            
            for field in outgoing_fields:
                if field in body:
                    mes[field] = body[field]
            
            if topic: #for metric_reducer is empty
                kafka_key = body.get(fld.kafka_key)
                self.send_message(header, mes, topic=topic, message_type=MessageType.DATA.value, kafka_key=kafka_key)
        else:
            pass
                
        self.send_monitor_metric(header, body)        
        
            
    def send_error(self, header, body, topic=None):
        '''Отправлялка ошибок'''
        if not header:
            header = Header('not-specified', MessageType.ERROR.value, self.client_id)        
        
        if not topic:
            topic = self.error_topic
        
        if not topic:
            return
        
        self.send_message(header, dict(error=body['error']), 
                          topic=topic, message_type=MessageType.ERROR.value)

        #ошибки тоже на мониторинг кидаем
        self.send_monitor_metric(header, body)                         
                        
    def send_monitor_metric(self, header, body, topics=[]):
        '''Отправлялка в мониторинг'''
        
        if not topics:
            topics = self.monitor_topics
        
        #монторинг понимает только ограниченный набор полей на метрики
        metric_body = {field:body[field] for field in self.metric_fields if field in body}
            
        if metric_body:
            
            if fld.timestamp in body:
                metric_body[fld.timestamp] = body[fld.timestamp]
            
            for topic in topics:
                
                if topic == MONITOR_IN_TOPIC:
                    metric_body = {k:v for k,v in metric_body.items() 
                                   if v != fld.to_detector} #to_detector for detector only
                               
                self.send_message(header, metric_body, message_type=MessageType.DATA.value, 
                                  topic=topic)
                self.logger.info(f'! send metric with session_id = {header.session_id} to topic = {topic}: {metric_body}')
                

class Actor(MessageBus):
    
    def preprocess_msg(self, msg):        
        '''пробуем распарсить
           если дадим на выход None, в колбек ничего не попадет'''
        
        try:            
            message = Message.parse_mes(msg)
        except Exception as e:
            trace_msg = traceback.format_exc()
            error = '! Cannot parse message in {0!s}: {1!r}'.format(type(self).__name__, trace_msg)
            self.logger.error(error)
            print(msg)
            raise(error)
            self.send_error(None, {'error':error})
            return None 
                         
        if message.header.message_type != MessageType.DATA.value:
            return None
        
        return message
                         
    def batch_callback(self, body):
        #в наследуемом классе, body - тело месседжа, на выходе res_body
        raise Exception('batch_callback method is not defined')
           
            
            
    def close_all(self):
        self.store.stop = True
        
    def process_messages(self, messages):
        
        errors_batch = []
        res_batch = []


        messages_batch = [[message.header, message.body] for message in 
                          (self.preprocess_msg(mes) for mes in messages) if message]

        if not messages_batch:
            return False
        


        try:
            
            microbatch_size = len(messages_batch)
            start_time = time()
            
            headers, bodies = list(zip(*messages_batch))
            
            
            out_bodies = self.batch_callback(bodies)
            
            cur_time = time()
            microbatch_time = cur_time - start_time
            mean_per_mes_time = microbatch_time/microbatch_size   

            if not out_bodies:
                return
            
            for i, body in enumerate(out_bodies):
                header = headers[i]
                if 'error' in body:
                    errors_batch.append([header, body])
                else:
                    if self.monitor_topics:
                        #additional metrics
                        if fld.metrics not in body:
                            body[fld.metrics] = {}
                        body[fld.metrics]['microbatch_size'] = microbatch_size
                        body[fld.metrics]['mean_per_mes_time'] = mean_per_mes_time
                        body[fld.metrics]['message_delay_time'] = microbatch_time
                        body[fld.metrics]['start_compute_time'] = start_time
                        
                    res_batch.append([header, body])                    

                    
        except Exception as e:
            trace_msg = traceback.format_exc()
            error = "Exception in {0!s}: {1!r}".format(type(self).__name__, trace_msg)
            self.logger.error(error)
            errors_batch = [[header, {'error':error}] for header, body in  messages_batch]
            
                            
        for header, body in res_batch:                
            self.process_result(header, body)
                        
        for header, body in errors_batch:
            self.send_error(header, body)     
            
    def run(self):
        print(f'{self.client_id} started')
        
        #get_mes_batch = lambda msgs: [mes for msg_groups in msgs.values() for mes in msg_groups]
        
        try:

            for mes_batch in self.store.reader_generator(self.in_topics):
                if not mes_batch:
                    continue      
              
                _ = self.process_messages(mes_batch)                

                
                #if not messages_batch:
                #    continue
                #print(f'{self.in_topics} mes_batch len: {len(messages_batch)}')
                #self.process_messages(messages_batch)

        except KeyboardInterrupt:
            self.logger.warning("Aborted by user")

        except Exception as e:
            trace_msg = traceback.format_exc()
            self.logger.error("Exception: {0!r}".format(trace_msg)) 
            print(f'{trace_msg}')

        finally:
            print(f'{self.client_id} stopped')
            self.close_all()
            
            
class Detector(Actor):        
    """
    """
    def __init__(self, config_dict):
        super().__init__(config_dict)        
        assert(len(self.monitor_topics) == 1)
        
    
    def process_result(self, header, res_body):
        if res_body:
            self.send_data(header, res_body)        
        
        
class Mapper(Actor):        
    """Маппилка с поддержкой распараллеливания
    """

    def __init__(self, config_dict):
        super().__init__(config_dict) 

    def process_result(self, header, res_body):
        self.send_data(header, res_body)
        
Reducer = Mapper 

class Dispatcher(Actor):        
    """Dispatcher умеет читать поле dispatch и отправлять сообщение в нужный id по топику 
    """

    def __init__(self, config_dict):
        super().__init__(config_dict) 

    def process_result(self, header, res_body):
        dest_id = res_body.pop('destination', None)
        
        if not dest_id:
            error = "Exception in {0!s}: {1}".format(type(self).__name__, 'no "destination" field in result')            
            self.send_error(header, {'error':error})
            return
        
        try:
            topic = self.out_topic_dict[dest_id]
            
        except:
            error = 'Exception in {0!s}: no topic in topic_dict for destination "{1}"'.format(type(self).__name__, dest_id)
            self.send_error(header, {'error':error})
            return
        
        self.send_data(header, res_body, topic=topic)